/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace MediaLab.TopSurf
{
    public partial class CompareForm : Form
    {
		// maintain a list to keep the bitmaps that we will show in the list view
		// Note: it may seem that keeping the bitmaplist is redundant,
		//       since bitmaps can be stored directly into the imagelist.
		//       however if we do so, for some reason C#'s memory use
		//       explodes and causes an 'out of memory' exception to be
		//       thrown upon closure of the form by the user, when
		//       disposing of the bitmaps. by disposing them from the
		//       bitmaplist instead everything works as it should.
		List<Bitmap> m_bitmaplist;
		public CompareForm(List<TOPSURF_DISTANCE> distance)
        {
            InitializeComponent();
			//  init the drawing size
			const int imagedim = 125; // this is the maximum size that is supported by a list view, as the width cannot be larger than 256 pixels
			int width = imagedim * 2 + 5;
			int height = imagedim;
			float aspect, x, y, w, h;
			// create the lists holding the list view images
			m_bitmaplist = new List<Bitmap>(distance.Count);
			compareList.LargeImageList = new ImageList();
			compareList.LargeImageList.ColorDepth = ColorDepth.Depth32Bit;
			compareList.LargeImageList.ImageSize = new Size(width, height);
			// convert the distance objects to list box entries
			if (distance == null || distance.Count == 0)
				throw new ArgumentException();
			foreach (TOPSURF_DISTANCE td in distance)
			{
				Bitmap bitmap = new Bitmap(width, height, PixelFormat.Format32bppArgb);
				using (Graphics g = Graphics.FromImage(bitmap))
				{
					g.Clear(Color.Transparent);
					// place both images aligned with their bottoms, but in the
					// center of their allocated space
					aspect = (float)td.Image1.Width / (float)td.Image1.Height;
					if (aspect > 1.0f)
					{
						w = imagedim;
						h = imagedim / aspect;
						x = 0;
						y = imagedim - h;
						
					}
					else
					{
						w = imagedim * aspect;
						h = imagedim;
						x = (imagedim - w) / 2;
						y = 0;
					}
					g.DrawImage(td.Image1, x, y, w, h);
					aspect = (float)td.Image2.Width / (float)td.Image2.Height;
					if (aspect > 1.0f)
					{
						w = imagedim;
						h = imagedim / aspect;
						x = imagedim + 5;
						y = imagedim - h;

					}
					else
					{
						w = imagedim * aspect;
						h = imagedim;
						x = imagedim + 5 + (imagedim - w) / 2;
						y = 0;
					}
					g.DrawImage(td.Image2, x, y, w, h);
				}
				// add the image to the list
				m_bitmaplist.Add(bitmap);
				compareList.LargeImageList.Images.Add(bitmap);
				// add the item to the list
				ListViewItem item = new ListViewItem(String.Format("{0:0.00}", td.Distance), compareList.Items.Count);
				compareList.Items.Add(item);
			}
			compareList.Location = new Point(0, 0);
			// set the size of the dialog
			// Note: setting the size causes the CompareForm_ClientSizeChanged function
			//       to be called, which places the compareList at the correct location
			this.ClientSize = new Size(compareList.Items[0].Bounds.Width + 20, compareList.Items.Count > 4 ? (compareList.Items[0].Bounds.Height + 15) * 4 : (compareList.Items[0].Bounds.Height + 15) * compareList.Items.Count);
			this.MinimumSize = this.Size;
        }

		private void CompareForm_ClientSizeChanged(object sender, EventArgs e)
		{
			// unless the form is minimized, the client size should always be
			// larger than zero
			if (this.ClientSize.Width >= 0 && this.ClientSize.Height >= 0)
				compareList.ClientSize = this.ClientSize;
		}

		private void CompareForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			compareList.LargeImageList.Images.Clear();
			foreach (Bitmap bitmap in m_bitmaplist)
				bitmap.Dispose();
		}
	}
}