﻿/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using MediaLab.Shell;
using MediaLab.FileBrowser;

namespace MediaLab.TopSurf
{
	public partial class ExtractDescriptorsForm : Form
	{
		// list of paths that point directly at images, or at
		// directories that contain images
		List<string> m_paths;
		public List<string> Paths { get { return m_paths; } }
		// number of images at which we will warn the user that
		// the extracted descriptors will not be automatically
		// visualized
		int m_maxVisualize;
		// number of images that we can subtract the descriptors of
		int m_maxExtract;
		// the windows explorer-like file browser
		ShellBrowser m_shellBrowser;
		TreeNode m_desktopNode, m_computerNode, m_selectedNode;
		ShellItem m_selectedItem;
		public enum SpecialFolders : uint
		{
			UserProfiles = 0x003e,
			DesktopDir_All = 0x0019,
			ApplicationData_All = 0x0023,
			MyDocuments_All = 0x002e,
			MyFavorites_All = 0x001f,
			MyMusic_All = 0x0035,
			MyPictures_All = 0x0036,
			StartMenu_All = 0x0016,
			MyVideos_All = 0x0037,
			Desktop = 0x0000,
			DekstopDir = 0x0010,
			MyComputer = 0x0011,
			MyFavorites = 0x0006,
			ApplicationData = 0x001c,
			MyDocuments = 0x000c,
			MyMusic = 0x000d,
			MyPictures = 0x0027,
			MyVideos = 0x000e,
			MyNetworkPlaces = 0x0012,
			MyDocumentsDir = 0x0005,
			StartMenu = 0x000b,
			ControlPanel = 0x0003,
			Printers = 0x0004,
			ProgramFiles = 0x0026,
			SendTo = 0x0009,
			System = 0x0025,
			Windows = 0x0024,
			RecycleBin = 0x000a
		}

		public ExtractDescriptorsForm(int maxVisualize, int maxExtract)
		{
			InitializeComponent();
			// prepare path list
			m_paths = new List<string>();
			// set the size of the dialog
			// Note: this calls ExtractDescriptorsForm_ClientSizeChanged, in which
			//       we position all other items
			this.ClientSize = new Size(750, 650);
			this.MinimumSize = this.ClientSize;
			// set maximum counts
			m_maxVisualize = maxVisualize;
			m_maxExtract = maxExtract;
			// set the bitmap and labels
			statusBox.BackgroundImage = null;
			statusBox.ClientSize = new Size(16, 16);
			statusBox.Visible = false;
			statusLabel.Text = "";
			statusLabel.Visible = false;
			countLabel.Text = "";
			countLabel.Visible = true;
			// disable the ok button until one or more images are added
			okButton.Enabled = false;
			// setup the file browser
			m_shellBrowser = new ShellBrowser();
			m_desktopNode = new TreeNode(m_shellBrowser.DesktopItem.Text, m_shellBrowser.DesktopItem.ImageIndex, m_shellBrowser.DesktopItem.SelectedImageIndex);
			m_desktopNode.Tag = m_shellBrowser.DesktopItem;
			m_desktopNode.Name = m_desktopNode.Text;
			folderView.Nodes.Add(m_desktopNode);
			m_selectedNode = m_desktopNode;
			m_selectedItem = null;
			// expand the desktop node and add all children
			m_shellBrowser.DesktopItem.Expand(false, true, IntPtr.Zero);
			foreach (ShellItem desktopChildItem in m_shellBrowser.DesktopItem.SubFolders)
			{
				TreeNode desktopChildNode = new TreeNode(desktopChildItem.Text, desktopChildItem.ImageIndex, desktopChildItem.SelectedImageIndex);
				desktopChildNode.Tag = desktopChildItem;
				desktopChildNode.Name = desktopChildNode.Text;
				if (desktopChildNode.Text == m_shellBrowser.MyComputerName)
				{
					m_computerNode = desktopChildNode;
					desktopChildItem.Expand(false, true, IntPtr.Zero);
					foreach (ShellItem computerChildItem in desktopChildItem.SubFolders)
					{
						TreeNode computerChildNode = new TreeNode(computerChildItem.Text, computerChildItem.ImageIndex, computerChildItem.SelectedImageIndex);
						computerChildNode.Tag = computerChildItem;
						computerChildNode.Name = computerChildNode.Text;
						if (computerChildItem.HasSubfolder)
							computerChildNode.Nodes.Add(string.Empty);
						desktopChildNode.Nodes.Add(computerChildNode);
					}
				}
				else if (desktopChildItem.HasSubfolder)
					desktopChildNode.Nodes.Add(string.Empty);
				m_desktopNode.Nodes.Add(desktopChildNode);
			}
			// add event handlers
			folderView.SetSorting(true);
			// navigate to the current directory
			// Note: it may occur that the current directory is an alias to a networked folder,
			//        e.g. the my documents folder at a university may point to a remote location,
			//        and in this case SelectPath will crash. in this event, select and expand
			//        the first node, which happens to be the desktop node.
			try	{ SelectPath(Directory.GetCurrentDirectory(), true); }
			catch { SelectPath(SpecialFolders.Desktop, true); }
			folderView.Focus();
		}

		private void ExtractDescriptorsForm_ClientSizeChanged(object sender, EventArgs e)
		{
			// check if we're not minimizing
			if (this.ClientSize.Width <= 0 || this.ClientSize.Height <= 0)
				return;
			// active area we want to paint on
			int areaWidth = this.ClientSize.Width - 10;
			int areaHeight = this.ClientSize.Height - 10;
			// add the info label to the top
			int labelHeight = infoLabel.ClientSize.Height;
			infoLabel.Location = new Point(5, 5);
			// add the status box and label to the bottom-left
			int iconWidth = statusBox.ClientSize.Width;
			int iconHeight = statusBox.ClientSize.Height;
			statusBox.Location = new Point(10, areaHeight - iconHeight + 2);
			statusLabel.Location = new Point(iconWidth + 5, areaHeight - (labelHeight + iconHeight) / 2);
			// add the ok and cancel buttons to the bottom-center
			int buttonWidth = okButton.ClientSize.Width;
			int buttonHeight = okButton.ClientSize.Height;
			okButton.Location = new Point((areaWidth - buttonWidth - 5) / 2, areaHeight - buttonHeight);
			cancelButton.Location = new Point(okButton.Location.X + buttonWidth + 5, areaHeight - buttonHeight);
			// add the count label to the bottom-right
			countLabel.Location = new Point(cancelButton.Location.X + buttonWidth + 5, areaHeight - (labelHeight + iconHeight) / 2);
			// keep the width of the folder view and the height of the
			// data box the same, while stretching the rest
			int splitHeight = areaHeight - (labelHeight + buttonHeight + 10);
			splitContainer.Location = new Point(5, labelHeight + 10);
			splitContainer.ClientSize = new Size(areaWidth, splitHeight);
			int dataWidth = areaWidth;
			int dataHeight = dataBox.ClientSize.Height;
			explorerContainer.ClientSize = new Size(areaWidth, splitHeight - dataHeight);
			int folderWidth = folderView.ClientSize.Width;
			int folderHeight = splitHeight - dataHeight;
			folderView.ClientSize = new Size(folderWidth, folderHeight);
			int fileWidth = areaWidth - folderWidth;
			int fileHeight = splitHeight - dataHeight;
			folderView.ClientSize = new Size(folderWidth, folderHeight);
		}

		private void ExtractDescriptorsForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			// release resources
			if (statusBox.BackgroundImage != null)
				statusBox.BackgroundImage.Dispose();
		}

		private string ConvertPath(string path)
		{
			if (string.IsNullOrEmpty(path))
				return path;
			string newPath = path.Trim();
			if (newPath.StartsWith(string.Format(@"{0}\", m_shellBrowser.MyComputerName), StringComparison.CurrentCulture) && newPath.Length > 12)
				newPath = newPath.Substring(path.IndexOf('\\') + 1);
			if (!newPath.EndsWith(@":\") && newPath.EndsWith(@"\"))
				newPath = newPath.Substring(0, newPath.Length - 1);
			if (newPath.EndsWith(@"\"))
				newPath = newPath.Substring(0, newPath.Length - 1);
			return newPath;
		}

		private bool PathExists(string path)
		{
			string realPath = ConvertPath(path);
			if (string.IsNullOrEmpty(realPath))
				return false;
			else if (string.Compare(path, "desktop", true) == 0)
				return true;
			string[] pathParts = realPath.Split('\\');
			for (int i = 0; i < pathParts.Length; i++)
			{
				bool found = false;
				if (m_shellBrowser.DesktopItem.SubFolders.Contains(pathParts[i]))
				{
					pathParts[i] = ShellItem.GetRealPath(m_shellBrowser.DesktopItem.SubFolders[pathParts[i]]);
					found = true;
				}
				else
				{
					ShellItem myComp = m_shellBrowser.DesktopItem.SubFolders[m_shellBrowser.MyComputerName];
					if (myComp.SubFolders.Contains(pathParts[i]))
					{
						pathParts[i] = ShellItem.GetRealPath(myComp.SubFolders[pathParts[i]]);
						found = true;
					}
				}
				if (!found)
					break;
			}
			realPath = string.Join("\\", pathParts);
			if (realPath.EndsWith(":"))
				realPath += "\\";
			ShellAPI.SHFILEINFO info = new ShellAPI.SHFILEINFO();
			IntPtr ptr = ShellAPI.SHGetFileInfo(realPath, 0, ref info, ShellAPI.cbFileInfo, ShellAPI.SHGFI.DISPLAYNAME);
			bool exists = (ptr != IntPtr.Zero);
			Marshal.FreeCoTaskMem(ptr);
			return exists;
		}

		public TreeNode SelectPath(SpecialFolders specialFolder, bool expandNode)
		{
			StringBuilder path = new StringBuilder(256);
			IntPtr pidl = IntPtr.Zero;
			if (specialFolder == SpecialFolders.Desktop)
				return SelectPath("Desktop", expandNode);
			else if (ShellAPI.SHGetFolderPath(IntPtr.Zero, (ShellAPI.CSIDL)specialFolder, IntPtr.Zero, ShellAPI.SHGFP.TYPE_CURRENT, path) == ShellAPI.S_OK)
			{
				path.Replace(m_shellBrowser.MyDocumentsPath, m_shellBrowser.MyDocumentsName);
				return SelectPath(path.ToString(), expandNode);
			}
			else
			{
				if (specialFolder == SpecialFolders.MyDocuments)
				{
					uint pchEaten = 0;
					ShellAPI.SFGAO pdwAttributes = 0;
					m_shellBrowser.DesktopItem.ShellFolder.ParseDisplayName(
						IntPtr.Zero,
						IntPtr.Zero,
						"::{450d8fba-ad25-11d0-98a8-0800361b1103}",
						ref pchEaten,
						out pidl,
						ref pdwAttributes);
				}
				else
					ShellAPI.SHGetSpecialFolderLocation(IntPtr.Zero, (ShellAPI.CSIDL)specialFolder, out pidl);
				if (pidl == IntPtr.Zero)
					return null;
				IntPtr strr = Marshal.AllocCoTaskMem(ShellAPI.MAX_PATH * 2 + 4);
				Marshal.WriteInt32(strr, 0, 0);
				StringBuilder buf = new StringBuilder(ShellAPI.MAX_PATH);
				if (m_shellBrowser.DesktopItem.ShellFolder.GetDisplayNameOf(pidl, ShellAPI.SHGNO.FORADDRESSBAR | ShellAPI.SHGNO.FORPARSING, strr) == ShellAPI.S_OK)
					ShellAPI.StrRetToBuf(strr, pidl, buf, ShellAPI.MAX_PATH);
				Marshal.FreeCoTaskMem(pidl);
				Marshal.FreeCoTaskMem(strr);
				if (string.IsNullOrEmpty(buf.ToString()))
					return null;
				return SelectPath(buf.ToString(), expandNode);
			}
		}

		public TreeNode SelectPath(string path, bool expandNode)
		{
			if (string.IsNullOrEmpty(path))
				return null;
			if (!PathExists(path))
				return null;
			folderView.BeginUpdate();
			string converted = ConvertPath(path);
			string[] pathParts = converted.Split('\\');
			TreeNode currentNode = null;
			if (string.Compare(pathParts[0], "desktop", true) == 0)
				currentNode = m_desktopNode;
			else if (m_desktopNode.Nodes.ContainsKey(pathParts[0]))
			{
				currentNode = m_desktopNode.Nodes[pathParts[0]];
				ExpandTreeNode(currentNode, false);
			}
			else
			{
				if (string.Compare(pathParts[0], m_computerNode.Text, true) == 0)
					currentNode = m_computerNode;
				else
				{
					if (pathParts[0][pathParts[0].Length - 1] == ':')
						pathParts[0] += "\\";
					foreach (TreeNode node in m_computerNode.Nodes)
					{
						if (string.Compare(pathParts[0], ((ShellItem)node.Tag).Path, true) == 0)
						{
							currentNode = node;
							ExpandTreeNode(currentNode, false);
							break;
						}
					}
				}
			}
			if (currentNode == null)
			{
				folderView.EndUpdate();
				return null;
			}
			for (int i = 1; i < pathParts.Length; i++)
			{
				if (pathParts[i][pathParts[i].Length - 1] == ':')
					pathParts[i] += "\\";
				bool found = false;
				foreach (TreeNode child in currentNode.Nodes)
				{
					if (string.Compare(pathParts[i], child.Text, true) == 0)
					{
						currentNode = child;
						found = true;
						break;
					}
				}
				if (!found)
				{
					folderView.EndUpdate();
					return null;
				}
				ExpandTreeNode(currentNode, false);
			}
			if (expandNode)
				currentNode.Expand();
			folderView.SelectedNode = currentNode;
			folderView.EndUpdate();
			// highlight the selected node and make sure it is visible
			folderView.Select();
			folderView.SelectedNode.EnsureVisible();
			return currentNode;
		}

		public TreeNode SelectPath(ShellItem item, bool expandNode)
		{
			if (item == null)
				return null;
			ShellItem[] path = m_shellBrowser.GetPath(item);
			if (path == null)
				return null;
			TreeNode currentNode = m_desktopNode;
			folderView.BeginUpdate();
			for (int i = 1; i < path.Length; i++)
			{
				ExpandTreeNode(currentNode, false);
				foreach (TreeNode subNode in currentNode.Nodes)
				{
					if (path[i].Equals(subNode.Tag))
					{
						currentNode = subNode;
						break;
					}
				}
			}
			if (expandNode)
				currentNode.Expand();
			folderView.EndUpdate();
			folderView.SelectedNode = currentNode;
			// highlight the selected node and make sure it is visible
			folderView.Select();
			folderView.SelectedNode.EnsureVisible();
			return currentNode;
		}

		private bool SelectNewPath(ShellItem oldItem, ShellItem newItem)
		{
			Cursor.Current = Cursors.WaitCursor;
			if (oldItem != newItem && newItem.Expand(true, false, IntPtr.Zero))
			{
				m_shellBrowser.UpdateCondition.ContinueUpdate = false;
				fileView.BeginUpdate();
				fileView.Items.Clear();
				fileView.ClearSelections();
				if (oldItem != null)
					oldItem.Clear(true, false);
				m_selectedItem = newItem;
				List<ListViewItem> newListItemsArray = new List<ListViewItem>();
				// add all folders
				foreach (ShellItem si in newItem.SubFolders)
				{
					ListViewItem item = new ListViewItem(si.Text, si.ImageIndex);
					item.Name = si.Text;
					item.Tag = si;
					newListItemsArray.Add(item);
				}
				// only add images
				foreach (ShellItem si in newItem.SubFiles)
				{
					if (ImageHelper.IsImage(si.Text))
					{
						ListViewItem item = new ListViewItem(si.Text, si.ImageIndex);
						item.Name = si.Text;
						item.Tag = si;
						newListItemsArray.Add(item);
					}
				}
				fileView.SetSorting(true);
				fileView.Items.AddRange(newListItemsArray.ToArray());
				fileView.SetSorting(false);
				fileView.EndUpdate();
				Cursor.Current = Cursors.Default;
				return true;
			}
			else
			{
				Cursor.Current = Cursors.Default;
				return (oldItem == newItem);
			}
		}

		private bool ExpandTreeNode(TreeNode node, bool overwrite)
        {
            if (!overwrite && !(node.Nodes.Count == 1 && string.IsNullOrEmpty(node.Nodes[0].Text)))
				return true;
            ShellItem nodeItem = (ShellItem)node.Tag;
            m_shellBrowser.UpdateCondition.ContinueUpdate = false;
            if (!nodeItem.Expand(false, true, IntPtr.Zero))
				return false;
            folderView.BeginUpdate();
            node.Nodes.Clear();
            TreeNode[] newNodesArray = new TreeNode[nodeItem.SubFolders.Count];
            for (int i = 0; i < newNodesArray.Length; i++)
            {
                newNodesArray[i] = new TreeNode(nodeItem.SubFolders[i].Text, nodeItem.SubFolders[i].ImageIndex, nodeItem.SubFolders[i].SelectedImageIndex);
                newNodesArray[i].Tag = nodeItem.SubFolders[i];
                if (nodeItem.SubFolders[i].HasSubfolder)
                    newNodesArray[i].Nodes.Add(string.Empty);
                newNodesArray[i].Name = newNodesArray[i].Text;
            }
            node.Nodes.AddRange(newNodesArray);
            folderView.EndUpdate();
            return true;
        }

		void folderView_AfterSelect(object sender, TreeViewEventArgs e)
		{
			ShellItem oldItem = m_selectedItem;
			ShellItem newItem = e.Node.Tag as ShellItem;
			if (SelectNewPath(oldItem, newItem))
			{
				if (!ShellItem.Equals(oldItem, newItem))
					m_selectedNode = folderView.SelectedNode;
			}
			else
				SelectPath(oldItem, false);
		}

		void folderView_BeforeSelect(object sender, TreeViewCancelEventArgs e)
		{
			ShellItem nodeItem = (ShellItem)e.Node.Tag;
			if (e.Action == TreeViewAction.ByMouse && !e.Node.IsExpanded)
				e.Node.Expand();
			else
				ExpandTreeNode(e.Node, false);
		}

		void folderView_BeforeExpand(object sender, TreeViewCancelEventArgs e)
		{
			ShellItem nodeItem = (ShellItem)e.Node.Tag;
			Cursor.Current = Cursors.WaitCursor;
			e.Cancel = !ExpandTreeNode(e.Node, false);
			Cursor.Current = Cursors.Default;
			if (e.Cancel)
				e.Node.Nodes.Clear();
		}

		private void fileView_MouseDoubleClick(object sender, MouseEventArgs e)
		{
			ListViewHitTestInfo info = fileView.HitTest(e.Location);
			ShellItem si = (ShellItem)info.Item.Tag;
			string path = ShellItem.GetRealPath(si);
			if (ImageHelper.IsImage(path))
			{
				// add to data box
				AddImage(path);
			}
			else
				SelectPath(si, true);
		}

		private void folderView_ItemDrag(object sender, ItemDragEventArgs e)
		{
			// get the dragged node
			// Note: this is not necessarily the same as the currently
			//       selected node
			TreeNode node = (TreeNode)e.Item;
			// add its shell item to a list for easy processing
			List<ShellItem> list = new List<ShellItem>(1);
			list.Add((ShellItem)node.Tag);
			// add the list to the drag-drop operation
			DoDragDrop(list, DragDropEffects.Copy);
		}

		private void fileView_ItemDrag(object sender, ItemDragEventArgs e)
		{
			// add all selected items to a list for easy processing
			List<ShellItem> list = new List<ShellItem>(fileView.SelectedItems.Count);
			foreach (object obj in fileView.SelectedItems)
			{
				ListViewItem item = (ListViewItem)obj;
				list.Add((ShellItem)item.Tag);
			}
			DoDragDrop(list, DragDropEffects.Copy);
		}

		private void dataBox_DragEnter(object sender, DragEventArgs e)
		{
			e.Effect = DragDropEffects.Copy;
		}

		private void dataBox_DragDrop(object sender, DragEventArgs e)
		{
			if (e.Data.GetDataPresent(typeof(List<ShellItem>)))
			{
				List<ShellItem> list = (List<ShellItem>)e.Data.GetData(typeof(List<ShellItem>));
				foreach (ShellItem si in list)
				{
					string path = ShellItem.GetRealPath(si);
					if (ImageHelper.IsImage(path))
						AddImage(path);
					else
						AddDirectory(path);
				}
			}
		}

		private void dataBox_KeyUp(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.Delete || e.KeyValue == 8) // 8 = backspace key
			{
				// delete the selected items
				while (dataBox.SelectedItems.Count > 0)
				{
					string path = (string)dataBox.SelectedItems[0];
					if (ImageHelper.IsImage(path))
						RemoveImage(path);
					else
						RemoveDirectory(path);
				}
			}
			else if (e.KeyCode == Keys.A && e.Control)
			{
				// select all items
				for (int i = 0; i < dataBox.Items.Count; i++)
					dataBox.SetSelected(i, true);
			}
		}

		private void AddImage(string path)
		{
			HideStatus();
			if (m_paths.Contains(path))
			{
				ShowInfo("The image is already included.");
				return;
			}
			if (m_paths.Count == m_maxExtract)
			{
				ShowError("A maximum of " + m_maxExtract + " images is supported. Image not added.");
				return;
			}
			m_paths.Add(path);
			if (m_paths.Count > m_maxVisualize)
				ShowWarning("The descriptors will not be visualized.");
			dataBox.Items.Add(path);
			UpdateCount();
		}

		private void RemoveImage(string path)
		{
			m_paths.Remove(path);
			if (m_paths.Count <= m_maxVisualize)
				HideStatus();
			dataBox.Items.Remove(path);
			UpdateCount();
		}

		private bool RecurseDirectory(string path, List<string> paths)
		{
			if (m_paths.Count + paths.Count > m_maxExtract)
				return false;
			try
			{
				string[] files = Directory.GetFiles(path);
				foreach (string file in files)
				{
					if (ImageHelper.IsImage(file) && !m_paths.Contains(file))
						paths.Add(file);
					if (m_paths.Count + paths.Count > m_maxExtract)
						return false;
				}
				string[] dirs = Directory.GetDirectories(path);
				foreach (string dir in dirs)
				{
					if (!RecurseDirectory(dir, paths))
						return false;
				}
			}
			catch
			{
				// probably this directory is not accessible, so simply continue
			}
			return true;
		}

		private void AddDirectory(string path)
		{
			HideStatus();
			if (dataBox.Items.Contains(path))
			{
				ShowInfo("The directory is already included.");
				return;
			}
			List<string> paths = new List<string>();
			if (!RecurseDirectory(path, paths))
			{
				ShowError("A maximum of " + m_maxExtract + " images is supported. Directory not added.");
				return;
			}
			m_paths.AddRange(paths);
			if (m_paths.Count > m_maxVisualize)
				ShowWarning("The descriptors will not be visualized.");
			dataBox.Items.Add(path);
			UpdateCount();
		}

		private void RemoveDirectory(string path)
		{
			m_paths.RemoveAll(x => x.StartsWith(path));
			if (m_paths.Count <= m_maxVisualize)
				HideStatus();
			dataBox.Items.Remove(path);
			UpdateCount();
		}

		private void HideStatus()
		{
			statusBox.Visible = false;
			statusLabel.Visible = false;
		}

		private void ShowInfo(string text)
		{
			if (statusBox.BackgroundImage != null)
				statusBox.BackgroundImage.Dispose();
			statusBox.BackgroundImage = SystemIcons.Information.ToBitmap();
			statusBox.Visible = true;
			statusLabel.Text = text;
			statusLabel.Visible = true;
		}

		private void ShowWarning(string text)
		{
			if (statusBox.BackgroundImage != null)
				statusBox.BackgroundImage.Dispose();
			statusBox.BackgroundImage = SystemIcons.Warning.ToBitmap();
			statusBox.Visible = true;
			statusLabel.Text = text;
			statusLabel.Visible = true;
		}

		private void ShowError(string text)
		{
			if (statusBox.BackgroundImage != null)
				statusBox.BackgroundImage.Dispose();
			statusBox.BackgroundImage = SystemIcons.Error.ToBitmap();
			statusBox.Visible = true;
			statusLabel.Text = text;
			statusLabel.Visible = true;
		}

		private void UpdateCount()
		{
			if (m_paths.Count > 0)
			{
				if (m_paths.Count == 1)
					countLabel.Text = "The descriptor of 1 image will be extracted.";
				else
					countLabel.Text = "The descriptors of " + m_paths.Count + " images will be extracted.";
				okButton.Enabled = true;
			}
			else
			{
				countLabel.Text = "";
				okButton.Enabled = false;
			}
		}
	}
}
