﻿/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Linq;
using System.Text;
using System.Windows.Forms;

namespace MediaLab.TopSurf
{
	public partial class GUIForm : Form
	{
		// topsurf object
		TOPSURF_DLL m_topsurf;
		// images and their extracted descriptor
		List<TOPSURF_GUIDATA> m_data;
		// visualize forms
		TOPSURF_FORM m_visualize;
		// compare forms
		TOPSURF_FORM m_compare;
		// maximum number of descriptors we support
		const int m_maxVisualize = 10;
		const int m_maxExtract = 1000;
		// remember the last used directories
		string m_lastDictionaryDirectory;
		string m_lastImageDirectory;
		string m_lastDescriptorDirectory;
		// prevent application from being closed while we access
		// the library
		bool m_preventClose = false;

		public GUIForm()
		{
			InitializeComponent();
			// enable the initialize menu item and disable all other menu items
			UpdateLibraryMenu(true, false);
			UpdateDictionaryMenu(false, false, false);
			UpdateDescriptorMenu(false, false, false, false, false);
			UpdateWindowMenu(true, false, false);
			UpdateHelpMenu(true);
			// set the size of the dialog
			this.ClientSize = new Size(logoBox.BackgroundImage.Width, logoBox.BackgroundImage.Height + menuStrip.Height + statusStrip.Height);
			this.MinimumSize = this.Size;
			this.MaximumSize = this.Size;
			// prepare the progress bar
			statusProgressBar.Size = new Size(200, statusProgressBar.Height);
			statusProgressBar.Visible = false;
			// set the background image
			logoBox.Location = new Point(0, menuStrip.Height);
			logoBox.ClientSize = this.ClientSize;
			// set the logger
			// Note: the box is a bit difficult to size due to extra edges and
			//       needing whole number of lines to fit on it
			logBox.Location = new Point(5, menuStrip.Height + 5);
			logBox.ClientSize = new Size(this.ClientSize.Width - 14, this.ClientSize.Height - menuStrip.Height - statusProgressBar.Height - 10);
			logBox.Visible = false;
			// create topsurf object
			m_topsurf = new TOPSURF_DLL(logBox);
			// create topsurf data list
			m_data = new List<TOPSURF_GUIDATA>();
			// create visualize form list
			m_visualize = new TOPSURF_FORM();
			// create compare form list
			m_compare = new TOPSURF_FORM();
			// set the last used directories to the current directory
			m_lastDictionaryDirectory = Directory.GetCurrentDirectory();
			m_lastImageDirectory = Directory.GetCurrentDirectory();
			m_lastDescriptorDirectory = Directory.GetCurrentDirectory();
		}

		private void GUIForm_Shown(object sender, EventArgs e)
		{
			// check if the DLL exists
			if (TOPSURF_DLL.Exists())
			{
				// show initialize menu when the form is shown for the first time
				initializeLibraryMenuItem_Click(null, null);
			}
			else
			{
				// disable the menu items
				DisableLibraryMenu();
				DisableDictionaryMenu();
				DisableDescriptorMenu();
				DisableWindowMenu();
				DisableHelpMenu();
				// inform user
				MessageBox.Show("Could not find the TOP-SURF library", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
			}
		}

		private void GUIForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (m_preventClose)
			{
				e.Cancel = true;
				return;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			DisableWindowMenu();
			DisableHelpMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			statusProgressBar.Visible = true;
			// close all forms
			m_visualize.Dispose();
			m_compare.Dispose();
			// release the descriptors
			string error;
			for (int i = 0; i < m_data.Count; i++)
			{
				TOPSURF_GUIDATA td = m_data[i];
				// write status message
				logBox.WriteNormal("Releasing descriptor " + td.Name + "...");
				// Note: first release the C++ memory and only then remove
				//       the data from the list. if we do it the other way
				//       around, C# will have destroyed the correct C++
				//       pointer to the data, causing the appliction to
				//       delete memory that is either invalid or belongs
				//       to other objects.
				if (!m_topsurf.ReleaseDescriptor(td.Descriptor, out error))
				{
					// write error message
					logBox.WriteError("failed!");
					if (!string.IsNullOrEmpty(error))
						logBox.WriteError(error);
					ShowLog();
					// continue to release the remaining descriptors
					continue;
				}
			}
			// separately dispose of the data, just in case we had issues with
			// releasing the descriptors via the library above
			for (int i = 0; i < m_data.Count; i++)
			{
				TOPSURF_GUIDATA td = m_data[i];
				td.Dispose();
			}
			m_data.Clear();
			// terminate the topsurf library
			// Note: we can still call this without problems even if the
			//       library has not been initialized at all, e.g. the
			//       user starts the application and closes it immediately
			if (!m_topsurf.Terminate(out error))
			{
				// write error message
				logBox.WriteError("failed!");
				if (!string.IsNullOrEmpty(error))
					logBox.WriteError(error);
				ShowLog();
			}
			statusProgressBar.Visible = false;
			m_topsurf = null;
		}

		private void UpdateLibraryMenu(bool initializeMenuItem, bool terminateMenuItem)
		{
			initializeDescriptorMenuItem.Tag = initializeDescriptorMenuItem.Enabled;
			initializeDescriptorMenuItem.Enabled = initializeMenuItem;
			terminateDescriptorMenuItem.Tag = terminateDescriptorMenuItem.Enabled;
			terminateDescriptorMenuItem.Enabled = terminateMenuItem;
			menuStrip.Items[0].Tag = menuStrip.Items[0].Enabled;
			if (initializeMenuItem || terminateMenuItem)
				menuStrip.Items[0].Enabled = true;
			else
				menuStrip.Items[0].Enabled = false;
			menuStrip.Refresh();
		}

		private void DisableLibraryMenu()
		{
			initializeDescriptorMenuItem.Tag = initializeDescriptorMenuItem.Enabled;
			initializeDescriptorMenuItem.Enabled = false;
			terminateDescriptorMenuItem.Tag = terminateDescriptorMenuItem.Enabled;
			terminateDescriptorMenuItem.Enabled = false;
			menuStrip.Items[0].Tag = menuStrip.Items[0].Enabled;
			menuStrip.Items[0].Enabled = false;
			menuStrip.Refresh();
		}

		private void RestoreLibraryMenu()
		{
			initializeDescriptorMenuItem.Enabled = (bool)initializeDescriptorMenuItem.Tag;
			terminateDescriptorMenuItem.Enabled = (bool)terminateDescriptorMenuItem.Tag;
			menuStrip.Items[0].Enabled = (bool)menuStrip.Items[0].Tag;
			menuStrip.Refresh();
		}

		private void UpdateDictionaryMenu(bool loadMenuItem, bool createMenuItem, bool saveMenuItem)
		{
			loadDictionaryMenuItem.Tag = loadDictionaryMenuItem.Enabled;
			loadDictionaryMenuItem.Enabled = loadMenuItem;
			createDictionaryMenuItem.Tag = createDictionaryMenuItem.Enabled;
			createDictionaryMenuItem.Enabled = createMenuItem;
			saveDictionaryMenuItem.Tag = saveDictionaryMenuItem.Enabled;
			saveDictionaryMenuItem.Enabled = saveMenuItem;
			menuStrip.Items[1].Tag = menuStrip.Items[1].Enabled;
			if (loadMenuItem || createMenuItem || saveMenuItem)
				menuStrip.Items[1].Enabled = true;
			else
				menuStrip.Items[1].Enabled = false;
			menuStrip.Refresh();
		}

		private void DisableDictionaryMenu()
		{
			loadDictionaryMenuItem.Tag = loadDictionaryMenuItem.Enabled;
			loadDictionaryMenuItem.Enabled = false;
			createDictionaryMenuItem.Tag = createDictionaryMenuItem.Enabled;
			createDictionaryMenuItem.Enabled = false;
			saveDictionaryMenuItem.Tag = saveDictionaryMenuItem.Enabled;
			saveDictionaryMenuItem.Enabled = false;
			menuStrip.Items[1].Tag = menuStrip.Items[1].Enabled;
			menuStrip.Items[1].Enabled = false;
			menuStrip.Refresh();
		}

		private void RestoreDictionaryMenu()
		{
			loadDictionaryMenuItem.Enabled = (bool)loadDictionaryMenuItem.Tag;
			createDictionaryMenuItem.Enabled = (bool)createDictionaryMenuItem.Tag;
			saveDictionaryMenuItem.Enabled = (bool)saveDictionaryMenuItem.Tag;
			menuStrip.Items[1].Enabled = (bool)menuStrip.Items[1].Tag;
			menuStrip.Refresh();
		}

		private void UpdateDescriptorMenu(bool extractMenuItem, bool visualizeMenuItem, bool saveMenuItem, bool compareMenuItem, bool releaseMenuItem)
		{
			extractDescriptorsMenuItem.Tag = extractDescriptorsMenuItem.Enabled;
			extractDescriptorsMenuItem.Enabled = extractMenuItem;
			visualizeDescriptorsMenuItem.Tag = visualizeDescriptorsMenuItem.Enabled;
			visualizeDescriptorsMenuItem.Enabled = visualizeMenuItem;
			saveDescriptorsMenuItem.Tag = saveDescriptorsMenuItem.Enabled;
			saveDescriptorsMenuItem.Enabled = saveMenuItem;
			compareDescriptorsMenuItem.Tag = compareDescriptorsMenuItem.Enabled;
			compareDescriptorsMenuItem.Enabled = compareMenuItem;
			releaseDescriptorsMenuItem.Tag = releaseDescriptorsMenuItem.Enabled;
			releaseDescriptorsMenuItem.Enabled = releaseMenuItem;
			menuStrip.Items[2].Tag = menuStrip.Items[2].Enabled;
			if (extractMenuItem || visualizeMenuItem || saveMenuItem || compareMenuItem || releaseMenuItem)
				menuStrip.Items[2].Enabled = true;
			else
				menuStrip.Items[2].Enabled = false;
			menuStrip.Refresh();
		}

		private void DisableDescriptorMenu()
		{
			extractDescriptorsMenuItem.Tag = extractDescriptorsMenuItem.Enabled;
			extractDescriptorsMenuItem.Enabled = false;
			visualizeDescriptorsMenuItem.Tag = visualizeDescriptorsMenuItem.Enabled;
			visualizeDescriptorsMenuItem.Enabled = false;
			saveDescriptorsMenuItem.Tag = saveDescriptorsMenuItem.Enabled;
			saveDescriptorsMenuItem.Enabled = false;
			compareDescriptorsMenuItem.Tag = compareDescriptorsMenuItem.Enabled;
			compareDescriptorsMenuItem.Enabled = false;
			releaseDescriptorsMenuItem.Tag = releaseDescriptorsMenuItem.Enabled;
			releaseDescriptorsMenuItem.Enabled = false;
			menuStrip.Items[2].Tag = menuStrip.Items[2].Enabled;
			menuStrip.Items[2].Enabled = false;
			menuStrip.Refresh();
		}

		private void RestoreDescriptorMenu()
		{
			extractDescriptorsMenuItem.Enabled = (bool)extractDescriptorsMenuItem.Tag;
			visualizeDescriptorsMenuItem.Enabled = (bool)visualizeDescriptorsMenuItem.Tag;
			saveDescriptorsMenuItem.Enabled = (bool)saveDescriptorsMenuItem.Tag;
			compareDescriptorsMenuItem.Enabled = (bool)compareDescriptorsMenuItem.Tag;
			releaseDescriptorsMenuItem.Enabled = (bool)releaseDescriptorsMenuItem.Tag;
			menuStrip.Items[2].Enabled = (bool)menuStrip.Items[2].Tag;
			menuStrip.Refresh();
		}

		private void UpdateWindowMenu(bool viewMenuItem, bool visualizeMenuItem, bool compareMenuItem)
		{
			viewLogMenuItem.Tag = viewLogMenuItem.Enabled;
			viewLogMenuItem.Enabled = viewMenuItem;
			visualizeWindowsMenuItem.Tag = visualizeWindowsMenuItem.Enabled;
			visualizeWindowsMenuItem.Enabled = visualizeMenuItem;
			compareWindowsMenuItem.Tag = compareWindowsMenuItem.Enabled;
			compareWindowsMenuItem.Enabled = compareMenuItem;
			menuStrip.Items[3].Tag = menuStrip.Items[3].Enabled;
			if (viewMenuItem || visualizeMenuItem || compareMenuItem)
				menuStrip.Items[3].Enabled = true;
			else
				menuStrip.Items[3].Enabled = false;
			menuStrip.Refresh();
		}

		private void DisableWindowMenu()
		{
			viewLogMenuItem.Tag = viewLogMenuItem.Enabled;
			viewLogMenuItem.Enabled = false;
			visualizeWindowsMenuItem.Tag = visualizeWindowsMenuItem.Enabled;
			visualizeWindowsMenuItem.Enabled = false;
			compareWindowsMenuItem.Tag = compareWindowsMenuItem.Enabled;
			compareWindowsMenuItem.Enabled = false;
			menuStrip.Items[3].Tag = menuStrip.Items[3].Enabled;
			menuStrip.Items[3].Enabled = false;
			menuStrip.Refresh();
		}

		private void RestoreWindowMenu()
		{
			viewLogMenuItem.Enabled = (bool)viewLogMenuItem.Tag;
			visualizeWindowsMenuItem.Enabled = (bool)visualizeWindowsMenuItem.Tag;
			compareWindowsMenuItem.Enabled = (bool)compareWindowsMenuItem.Tag;
			menuStrip.Items[3].Enabled = (bool)menuStrip.Items[3].Tag;
			menuStrip.Refresh();
		}

		private void UpdateHelpMenu(bool aboutMenuItem)
		{
			aboutApplicationMenuItem.Tag = aboutApplicationMenuItem.Enabled;
			aboutApplicationMenuItem.Enabled = aboutMenuItem;
			menuStrip.Items[3].Tag = menuStrip.Items[3].Enabled;
			if (aboutMenuItem)
				menuStrip.Items[4].Enabled = true;
			else
				menuStrip.Items[4].Enabled = false;
			menuStrip.Refresh();
		}

		private void DisableHelpMenu()
		{
			aboutApplicationMenuItem.Tag = aboutApplicationMenuItem.Enabled;
			aboutApplicationMenuItem.Enabled = false;
			menuStrip.Items[4].Tag = menuStrip.Items[4].Enabled;
			menuStrip.Items[4].Enabled = false;
			menuStrip.Refresh();
		}

		private void RestoreHelpMenu()
		{
			aboutApplicationMenuItem.Enabled = (bool)aboutApplicationMenuItem.Tag;
			menuStrip.Items[4].Enabled = (bool)menuStrip.Items[4].Tag;
			menuStrip.Refresh();
		}

		private void initializeLibraryMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the default image size to
			// resize to and the number of elements to retain when extracting
			// a descriptor
			int imagedim, top;
			using (InitializeLibraryForm form = new InitializeLibraryForm())
			{
				if (form.ShowDialog() != DialogResult.OK)
					return;
				imagedim = form.ImageDim;
				top = form.TopElements;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			DisableWindowMenu();
			DisableHelpMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			statusProgressBar.Visible = true;
			// write status message
			logBox.WriteNormal("Initializing TOP-SURF library...");
			// initialize the topsurf library
			string error;
			if (!m_topsurf.Initialize(imagedim, top, out error))
			{
				// restore the menu items
				RestoreLibraryMenu();
				RestoreDictionaryMenu();
				RestoreDescriptorMenu();
				RestoreWindowMenu();
				RestoreHelpMenu();
				statusProgressBar.Visible = false;
				// write error message
				logBox.WriteError("failed!");
				if (!string.IsNullOrEmpty(error))
					logBox.WriteError(error);
				ShowLog();
				m_preventClose = false;
				return;
			}
			// update/restore the menu items
			UpdateLibraryMenu(false, true);
			UpdateDictionaryMenu(true, true, false);
			UpdateDescriptorMenu(false, false, false, false, false);
			RestoreWindowMenu();
			RestoreHelpMenu();
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void terminateLibraryMenuItem_Click(object sender, EventArgs e)
		{
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			DisableWindowMenu();
			DisableHelpMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			statusProgressBar.Visible = true;
			// write status message
			logBox.WriteNormal("Terminating TOP-SURF library...");
			// terminate the topsurf library
			string error;
			if (!m_topsurf.Terminate(out error))
			{
				// restore the menu items
				RestoreLibraryMenu();
				RestoreDictionaryMenu();
				RestoreDescriptorMenu();
				RestoreWindowMenu();
				RestoreHelpMenu();
				statusProgressBar.Visible = false;
				// write error message
				logBox.WriteError("failed!");
				if (!string.IsNullOrEmpty(error))
					logBox.WriteError(error);
				ShowLog();
				m_preventClose = false;
				return;
			}
			// release resources
			for (int i = 0; i < m_data.Count; i++)
			{
				TOPSURF_GUIDATA td = m_data[i];
				td.Dispose();
			}
			m_data.Clear();
			// update/restore the menu items
			UpdateLibraryMenu(true, false);
			UpdateDictionaryMenu(false, false, false);
			UpdateDescriptorMenu(false, false, false, false, false);
			RestoreWindowMenu();
			RestoreHelpMenu();
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void loadDictionaryMenuItem_Click(object sender, EventArgs e)
		{
			// check if all descriptors have been released
			if (m_data.Count > 0)
			{
				if (MessageBox.Show("When you load a new dictionary, all currently extracted descriptors will be released.", "Warning", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning) != DialogResult.OK)
					return;
				// release all descriptors
				// Note: we must make a copy of the list, otherwise we loop through and delete
				//       from the same list in the function
				ReleaseDescriptors(new List<TOPSURF_GUIDATA>(m_data));
			}
			// pop up dialog box to ask the user for the directory where to
			// load the dictionary from
			// Note: let it use the last directory that the user selected
			string currentDirectory = Directory.GetCurrentDirectory();
			try { Directory.SetCurrentDirectory(m_lastDictionaryDirectory); }
			catch { };;
			string path;
			using (LoadDictionaryForm form = new LoadDictionaryForm())
			{
				// show the form
				DialogResult dialogResult = form.ShowDialog();
				// restore the old directory
				try { Directory.SetCurrentDirectory(currentDirectory); }
				catch { };
				if (dialogResult != DialogResult.OK)
					return;
				// get the user's choices
				m_lastDictionaryDirectory = form.Directory;
				path = form.Directory;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			statusProgressBar.Visible = true;
			// write status message
			logBox.WriteNormal("Loading dictionary...");
			// load the dictionary
			string error;
			if (!m_topsurf.LoadDictionary(path, out error))
			{
				// restore the menu items
				RestoreLibraryMenu();
				RestoreDictionaryMenu();
				RestoreDescriptorMenu();
				statusProgressBar.Visible = false;
				// write error message
				logBox.WriteError("failed!");
				if (!string.IsNullOrEmpty(error))
					logBox.WriteError(error);
				ShowLog();
				m_preventClose = false;
				return;
			}
			// update/restore the menu items
			UpdateLibraryMenu(false, true);
			UpdateDictionaryMenu(true, true, true);
			UpdateDescriptorMenu(true, false, false, false, false);
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void createDictionaryMenuItem_Click(object sender, EventArgs e)
		{
			// check if all descriptors have been released
			if (m_data.Count > 0)
			{
				if (MessageBox.Show("When you load a new dictionary, all currently extracted descriptors will be released.", "Warning", MessageBoxButtons.OKCancel, MessageBoxIcon.Warning) != DialogResult.OK)
					return;
				// release all descriptors
				// Note: we must make a copy of the list, otherwise we loop through and delete
				//       from the same list in the function
				ReleaseDescriptors(new List<TOPSURF_GUIDATA>(m_data));
			}
			// pop up dialog box to ask the user for the settings with which
			// to create the dictionary
			string currentDirectory = Directory.GetCurrentDirectory();
			try { Directory.SetCurrentDirectory(m_lastImageDirectory); }
			catch { };;
			string path;
			int clusters, knn, iterations, points;
			using (CreateDictionaryForm form = new CreateDictionaryForm())
			{
				// show the form
				DialogResult dialogResult = form.ShowDialog();
				// restore the old directory
				try { Directory.SetCurrentDirectory(currentDirectory); }
				catch { };
				if (dialogResult != DialogResult.OK)
					return;
				// get the user's choices
				m_lastImageDirectory = form.ImageDirectory;
				path = form.ImageDirectory;
				clusters = form.Clusters;
				knn = form.KNN;
				iterations = form.Iterations;
				points = form.Points;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			statusProgressBar.Visible = true;
			// write status message
			logBox.WriteNormal("Creating dictionary...");
			// create the dictionary
			string error;
			if (!m_topsurf.CreateDictionary(path, clusters, knn, iterations, points, out error))
			{
				// restore the menu items
				RestoreLibraryMenu();
				RestoreDictionaryMenu();
				RestoreDescriptorMenu();
				statusProgressBar.Visible = false;
				// write error message
				logBox.WriteError("failed!");
				if (!string.IsNullOrEmpty(error))
					logBox.WriteError(error);
				ShowLog();
				m_preventClose = false;
				return;
			}
			// update/restore the menu items
			UpdateLibraryMenu(false, true);
			UpdateDictionaryMenu(true, true, true);
			UpdateDescriptorMenu(true, false, false, false, false);
			statusProgressBar.Visible = false;
			// popup the save dialog box to save the newly created dictionary
			saveDictionaryMenuItem_Click(null, null);
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void saveDictionaryMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the directory where to
			// save the dictionary to
			string currentDirectory = Directory.GetCurrentDirectory();
			try { Directory.SetCurrentDirectory(m_lastDictionaryDirectory); }
			catch { };;
			string path;
			using (SaveDictionaryForm form = new SaveDictionaryForm())
			{
				// show the form
				DialogResult dialogResult = form.ShowDialog();
				// restore the old directory
				try { Directory.SetCurrentDirectory(currentDirectory); }
				catch { };
				if (dialogResult != DialogResult.OK)
					return;
				// get the user's choices
				m_lastDictionaryDirectory = form.Directory;
				path = form.Directory;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			statusProgressBar.Visible = true;
			// write status message
			logBox.WriteNormal("Saving dictionary...");
			// save the dictionary
			string error;
			if (!m_topsurf.SaveDictionary(path, out error))
			{
				// restore the menu items
				RestoreLibraryMenu();
				RestoreDictionaryMenu();
				RestoreDescriptorMenu();
				statusProgressBar.Visible = false;
				// write error message
				logBox.WriteError("failed!");
				if (!string.IsNullOrEmpty(error))
					logBox.WriteError(error);
				ShowLog();
				m_preventClose = false;
				return;
			}
			// restore the menu items
			RestoreLibraryMenu();
			RestoreDictionaryMenu();
			RestoreDescriptorMenu();
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void Bitmap2Pixels(Bitmap bitmap, out byte[] pixels, out int width, out int height)
		{
			// Note: by default request in 24-bit RGB. the pixels that are returned are
			//       ordered by BGR and not RGB
			width = bitmap.Width;
			height = bitmap.Height;
			pixels = new byte[width * height * 3];
			BitmapData imageData = bitmap.LockBits(new Rectangle(0, 0, width, height),
					ImageLockMode.ReadOnly, PixelFormat.Format24bppRgb);
			unsafe
			{
				byte* p;
				for (int y = 0, i = 0; y < height; y++)
				{
					p = (byte*)imageData.Scan0 + y * imageData.Stride;
					for (int x = 0; x < width; x++, i += 3, p += 3)
					{
						pixels[i] = p[2];
						pixels[i + 1] = p[1];
						pixels[i + 2] = p[0];
					}
				}
			}
			bitmap.UnlockBits(imageData);
		}

		private void Pixels2Image(byte[] pixels, int width, int height, out Bitmap bitmap)
		{
			bitmap = new Bitmap(width, height);
			BitmapData visualizeData = bitmap.LockBits(new Rectangle(0, 0, width, height),
				ImageLockMode.WriteOnly, PixelFormat.Format24bppRgb);
			unsafe
			{
				byte* p;
				for (int y = 0, i = 0; y < height; y++)
				{
					p = (byte*)visualizeData.Scan0 + y * visualizeData.Stride;
					for (int x = 0; x < width; x++, i += 3, p += 3)
					{
						p[2] = pixels[i];
						p[1] = pixels[i + 1];
						p[0] = pixels[i + 2];
					}
				}
			}
			bitmap.UnlockBits(visualizeData);
		}

		private void Visualize(List<TOPSURF_GUIDATA> data)
		{
			foreach (TOPSURF_GUIDATA td in data)
			{
				// write status message
				logBox.WriteNormal("Visualizing descriptor " + td.Name + "...");
				// show the image form
				VisualizeForm f = new VisualizeForm(td);
				m_visualize.Add(f);
				f.Show();
			}
		}

		private void extractDescriptorMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the images from which
			// to extract the descriptors
			// Note: let it use the last directory that the user selected
			string currentDirectory = Directory.GetCurrentDirectory();
			try { Directory.SetCurrentDirectory(m_lastImageDirectory); }
			catch { };;
			List<String> paths;
			int maxExtract = m_maxExtract - m_data.Count;
			using (ExtractDescriptorsForm form = new ExtractDescriptorsForm(m_maxVisualize, maxExtract))
			{
				// show the form
				DialogResult dialogResult = form.ShowDialog();
				// restore the old directory
				try { Directory.SetCurrentDirectory(currentDirectory); }
				catch { };
				if (dialogResult != DialogResult.OK)
					return;
				// get the user's choices
				paths = form.Paths;
				m_lastImageDirectory = Path.GetDirectoryName(form.Paths[form.Paths.Count - 1]);
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Blocks;
			statusProgressBar.Minimum = 0;
			statusProgressBar.Maximum = paths.Count;
			statusProgressBar.Value = 0;
			statusProgressBar.Step = 1;
			statusProgressBar.Visible = true;
			// extract the descriptors of each image and generate their visualization bitmap
			List<TOPSURF_GUIDATA> data = new List<TOPSURF_GUIDATA>(paths.Count);
			foreach (string path in paths)
			{
				// write status message
				logBox.WriteNormal("Extracting descriptor " + Path.GetFileNameWithoutExtension(path) + "...");
				// load image
				Bitmap bitmap;
				try
				{
					bitmap = new Bitmap(path);
				}
				catch
				{
					logBox.WriteError(path + " is not a valid image");
					continue;
				}
				// convert the bitmap to a byte array
				int width;
				int height;
				byte[] pixels;
				Bitmap2Pixels(bitmap, out pixels, out width, out height);
				// extract the descriptor and visualize it
				TOPSURF_CXX descriptor;
				string error;
				bool success = false;
				if (m_topsurf.ExtractDescriptor(pixels, width, height, out descriptor, out error))
				{
					// we have used the original image to extract the descriptor, but to display
					// it we may have to resize it somewhat, and also to not use too much memory
					Bitmap image;
					if (bitmap.Width > 640 || bitmap.Height > 640)
					{
						float aspect = (float)bitmap.Width / (float)bitmap.Height;
						if (aspect > 1.0f)
							image = new Bitmap(bitmap, 640, (int)(640 / aspect));
						else
							image = new Bitmap(bitmap, (int)(640 * aspect), 640);
						bitmap.Dispose();
					}
					else
						image = bitmap;
					// convert the image to a byte array
					Bitmap2Pixels(image, out pixels, out width, out height);
					// visualize the descriptor
					if (m_topsurf.VisualizeDescriptor(pixels, width, height, descriptor, out error))
					{
						// convert the pixels back into an image
						Bitmap visualize;
						Pixels2Image(pixels, width, height, out visualize);
						// add the bitmaps and the descriptor to the data list
						data.Add(new TOPSURF_GUIDATA(path, image, visualize, descriptor));
						success = true;
					}
				}
				// check if everything happened successfully
				if (!success)
				{
					// release resources
					bitmap.Dispose();
					for (int i = 0; i < data.Count; i++)
						data[i].Dispose();
					data.Clear();
					// restore the menu items
					RestoreLibraryMenu();
					RestoreDictionaryMenu();
					RestoreDescriptorMenu();
					statusProgressBar.Visible = false;
					// write error message
					logBox.WriteError("failed!");
					if (!string.IsNullOrEmpty(error))
						logBox.WriteError(error);
					ShowLog();
					m_preventClose = false;
					return;
				}
				// update progress bar
				statusProgressBar.PerformStep();
			}
			// visualize all extracted descriptors if there are not too many
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			if (data.Count <= m_maxVisualize)
				Visualize(data);
			// add the new descriptors to the descriptor list
			m_data.AddRange(data);
			// update/restore the menu items
			// Note: the contents of the window menu will be updated when
			//       the user clicks on the menu title
			UpdateLibraryMenu(false, true);
			UpdateDictionaryMenu(true, true, true);
			if (m_data.Count == 1)
				UpdateDescriptorMenu(true, true, true, false, true); 
			else if (m_data.Count == m_maxExtract)
				UpdateDescriptorMenu(false, true, true, true, true);
			else
				UpdateDescriptorMenu(true, true, true, true, true);
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void visualizeDescriptorMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the descriptors to visualize
			List<TOPSURF_GUIDATA> data;
			using (VisualizeDescriptorsForm form = new VisualizeDescriptorsForm(m_data))
			{
				if (form.ShowDialog() != DialogResult.OK)
					return;
				data = form.Data;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Marquee;
			// show the visualizations
			Visualize(data);
			// restore the menu items
			// Note: the contents of the window menu will be updated when
			//       the user clicks on the menu title
			RestoreLibraryMenu();
			RestoreDictionaryMenu();
			RestoreDescriptorMenu();
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void saveDescriptorMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the descriptors to save
			// and where to save them
			string currentDirectory = Directory.GetCurrentDirectory();
			try { Directory.SetCurrentDirectory(m_lastDescriptorDirectory); }
			catch { };;
			List<TOPSURF_GUIDATA> data;
			string folder;
			using (SaveDescriptorsForm form = new SaveDescriptorsForm(m_data))
			{
				// show the form
				DialogResult dialogResult = form.ShowDialog();
				// restore the old directory
				try { Directory.SetCurrentDirectory(currentDirectory); }
				catch { };
				if (dialogResult != DialogResult.OK)
					return;
				// get the user's choices
				data = form.Data;
				folder = form.Folder;
				if (!string.IsNullOrEmpty(form.Folder))
					m_lastDescriptorDirectory = form.Folder;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Blocks;
			statusProgressBar.Minimum = 0;
			statusProgressBar.Maximum = data.Count;
			statusProgressBar.Value = 0;
			statusProgressBar.Step = 1;
			statusProgressBar.Visible = true;
			// save the descriptors
			for (int i = 0; i < data.Count; i++)
			{
				TOPSURF_GUIDATA td = data[i];
				// write status message
				logBox.WriteNormal("Saving descriptor " + td.Name + "...");
				// compose the path where to save to
				string path;
				if (string.IsNullOrEmpty(folder))
				{
					// same folder as the image
					path = td.Path;
				}
				else
				{
					// new folder
					path = Path.Combine(folder, td.Name);
				}
				// add the top extension
				path = Path.ChangeExtension(path, "top");
				// save the descriptor
				string error;
				if (!m_topsurf.SaveDescriptor(path, td.Descriptor, out error))
				{
					// restore the menu items
					RestoreLibraryMenu();
					RestoreDictionaryMenu();
					RestoreDescriptorMenu();
					statusProgressBar.Visible = false;
					// write error message
					logBox.WriteError("failed!");
					if (!string.IsNullOrEmpty(error))
						logBox.WriteError(error);
					ShowLog();
					m_preventClose = false;
					return;
				}
				// update progress bar
				statusProgressBar.PerformStep();
			}
			// restore the menu items
			RestoreLibraryMenu();
			RestoreDictionaryMenu();
			RestoreDescriptorMenu();
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void compareDescriptorsMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the descriptors to compare
			TOPSURF_GUIDATA data1;
			List<TOPSURF_GUIDATA> data2;
			TOPSURF_SIMILARITY similarity;
			using (CompareDescriptorsForm form = new CompareDescriptorsForm(m_data))
			{
				if (form.ShowDialog() != DialogResult.OK)
					return;
				data1 = form.Data1;
				data2 = form.Data2;
				similarity = form.Similarity;
			}
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Blocks;
			statusProgressBar.Minimum = 0;
			statusProgressBar.Maximum = data2.Count;
			statusProgressBar.Value = 0;
			statusProgressBar.Step = 1;
			statusProgressBar.Visible = true;
			// compare descriptors
			List<TOPSURF_DISTANCE> distance = new List<TOPSURF_DISTANCE>(data2.Count);
			for (int i = 0; i < data2.Count; i++)
			{
				// write status message
				logBox.WriteNormal("Comparing descriptor " + data1.Name + " with " + data2[i].Name + "...");					
				// compare the two descriptors
				float d;
				string error;
				if (!m_topsurf.CompareDescriptors(data1.Descriptor, data2[i].Descriptor, similarity, out d, out error))
				{
					// restore the menu items
					RestoreLibraryMenu();
					RestoreDictionaryMenu();
					RestoreDescriptorMenu();
					statusProgressBar.Visible = false;
					// write error message
					logBox.WriteError("failed!");
					if (!string.IsNullOrEmpty(error))
						logBox.WriteError(error);
					ShowLog();
					m_preventClose = false;
					return;
				}
				distance.Add(new TOPSURF_DISTANCE(data1.Image, data2[i].Image, d));
				// update progress bar
				statusProgressBar.PerformStep();
			}
			// sort the distance list
			distance.Sort();
			// show them in the compare dialog
			CompareForm f = new CompareForm(distance);
			m_compare.Add(f);
			f.Show();
			// restore the menu items
			RestoreLibraryMenu();
			RestoreDictionaryMenu();
			RestoreDescriptorMenu();
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void releaseDescriptorsMenuItem_Click(object sender, EventArgs e)
		{
			// pop up dialog box to ask the user for the descriptors to release
			List<TOPSURF_GUIDATA> data;
			using (ReleaseDescriptorsForm form = new ReleaseDescriptorsForm(m_data))
			{
				if (form.ShowDialog() != DialogResult.OK)
					return;
				data = form.Data;
			}
			// release the descriptors
			ReleaseDescriptors(data);
		}

		private void ReleaseDescriptors(List<TOPSURF_GUIDATA> data)
		{
			// prevent the user from closing the application
			m_preventClose = true;
			// disable the menu items
			// Note: except the window and help menus so the user can still
			//       have limited access to the interface
			DisableLibraryMenu();
			DisableDictionaryMenu();
			DisableDescriptorMenu();
			// show the progress bar
			statusProgressBar.Style = ProgressBarStyle.Blocks;
			statusProgressBar.Minimum = 0;
			statusProgressBar.Maximum = data.Count;
			statusProgressBar.Value = 0;
			statusProgressBar.Step = 1;
			statusProgressBar.Visible = true;
			// release the descriptors
			for (int i = 0; i < data.Count; i++)
			{
				TOPSURF_GUIDATA td = data[i];
				// write status message
				logBox.WriteNormal("Releasing descriptor " + td.Name + "...");
				// Note: first release the C++ memory and only then remove
				//       the data from the list. if we do it the other way
				//       around, C# will have destroyed the correct C++
				//       pointer to the data, causing the appliction to
				//       delete memory that is either invalid or belongs
				//       to other objects.
				string error;
				if (!m_topsurf.ReleaseDescriptor(td.Descriptor, out error))
				{
					// restore the menu items
					RestoreLibraryMenu();
					RestoreDictionaryMenu();
					RestoreDescriptorMenu();
					statusProgressBar.Visible = false;
					// write error message
					logBox.WriteError("failed!");
					if (!string.IsNullOrEmpty(error))
						logBox.WriteError(error);
					ShowLog();
					m_preventClose = false;
					return;
				}
				m_data.RemoveAll(x => x.Path.Equals(td.Path));
				td.Dispose();
				// update progress bar
				statusProgressBar.PerformStep();
			}
			// restore/update the menu items
			RestoreLibraryMenu();
			RestoreDictionaryMenu();
			if (m_data.Count == 0)
				UpdateDescriptorMenu(true, false, false, false, false);
			else if (m_data.Count == 1)
				UpdateDescriptorMenu(true, true, true, false, true);
			else
				UpdateDescriptorMenu(true, true, true, true, true);
			statusProgressBar.Visible = false;
			logBox.WriteNormal("done!");
			m_preventClose = false;
		}

		private void windowMenuItem_Click(object sender, EventArgs e)
		{
			// as the user might have closed one or more visualization
			// and/or compare windows, we have to check whether or not
			// we should enable or disable the close window menu options
			UpdateWindowMenu(true, m_visualize.Count > 0 ? true : false, m_compare.Count > 0 ? true : false);
		}

		private void viewLogMenuItem_Click(object sender, EventArgs e)
		{
			if (logBox.Visible)
				HideLog();
			else
				ShowLog();
		}

		private void HideLog()
		{
			logBox.Visible = false;
			logoBox.Visible = true;
			this.viewLogMenuItem.Text = "View Log";
		}
		
		private void ShowLog()
		{
			logBox.Visible = true;
			logoBox.Visible = false;
			this.viewLogMenuItem.Text = "Hide Log";
		}

		private void closeWindowsMenuItem_Click(object sender, EventArgs e)
		{
			m_visualize.CloseAll();
		}

		private void closeAllComparisonWindowsToolStripMenuItem_Click(object sender, EventArgs e)
		{
			m_compare.CloseAll();
		}

		private void aboutApplicationMenuItem_Click(object sender, EventArgs e)
		{
			using (AboutBox form = new AboutBox())
				form.ShowDialog();
		}
	}
}
