﻿/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Runtime.InteropServices;
using System.Text;

namespace MediaLab.TopSurf
{
	// the TOP-SURF descriptor as we get it from the C++ DLL
	public class TOPSURF_CXX
	{
		public IntPtr data;
		public int length;
	}
	
	// a TOP-SURF visual word location
	public class TOPSURF_LOCATION
	{
		public float x;
		public float y;
		public float scale;
		public float orientation;
	}

	// a TOP-SURF visual word
	public class TOPSURF_VISUALWORD
	{
		public int identifier;
		public float tf;
		public float idf;
		public TOPSURF_LOCATION[] location;
	}

	// a TOP-SURF descriptor
	public class TOPSURF_DESCRIPTOR
	{
		public float length;
		public TOPSURF_VISUALWORD[] visualword;

		// convert the C++ descriptor to a C# descriptor
		// Note: by default we get a pointer to the C++ descriptor, which we can convert to our C#
		//       descriptor type. in the future we might figure out how to directly use the C++
		//       descriptor without requiring the intermediate byte array and its conversion.
		public static TOPSURF_DESCRIPTOR ConvertDescriptor(TOPSURF_CXX cxx)
		{
			try
			{
				// copy the unmanaged array to a managed array
				byte[] b = new byte[cxx.length];
				Marshal.Copy(cxx.data, b, 0, cxx.length);
				// create a new descriptor and walk through the
				// array filling the data as we go
				int index = 0;
				TOPSURF_DESCRIPTOR descriptor = new TOPSURF_DESCRIPTOR();
				int visualwordCount = BitConverter.ToInt32(b, index);
				if (visualwordCount < 0)
					return null;
				else if (visualwordCount == 0)
				{
					descriptor.length = 0;
					descriptor.visualword = new TOPSURF_VISUALWORD[0];
				}
				else
				{
					index += sizeof(Int32);
					descriptor.length = BitConverter.ToSingle(b, index);
					index += sizeof(Single);
					descriptor.visualword = new TOPSURF_VISUALWORD[visualwordCount];
					for (int i = 0; i < visualwordCount; i++)
					{
						TOPSURF_VISUALWORD vw = new TOPSURF_VISUALWORD();
						vw.identifier = BitConverter.ToInt32(b, index);
						index += sizeof(Int32);
						vw.tf = BitConverter.ToSingle(b, index);
						index += sizeof(Single);
						vw.idf = BitConverter.ToSingle(b, index);
						index += sizeof(Single);
						int locationCount = BitConverter.ToInt32(b, index);
						index += sizeof(Int32);
						if (locationCount <= 0)
							return null;
						vw.location = new TOPSURF_LOCATION[locationCount];
						for (int j = 0; j < locationCount; j++)
						{
							TOPSURF_LOCATION loc = new TOPSURF_LOCATION();
							loc.x = BitConverter.ToSingle(b, index);
							index += sizeof(Single);
							loc.y = BitConverter.ToSingle(b, index);
							index += sizeof(Single);
							loc.orientation = BitConverter.ToSingle(b, index);
							index += sizeof(Single);
							loc.scale = BitConverter.ToSingle(b, index);
							index += sizeof(Single);
							vw.location[j] = loc;
						}
						descriptor.visualword[i] = vw;
					}
				}
				return descriptor;
			}
			catch (Exception ex)
			{
				Debug.WriteLine(ex.Message);
				return null;
			}
		}
	}

	// choice of similarity measure for comparing descriptors
	public enum TOPSURF_SIMILARITY
	{
		TOPSURF_COSINE,  // cosine normalized difference between the tf-idf scores of the visual words
		TOPSURF_ABSOLUTE // absolute difference between the td-idf scores of the visual words
	};
}
