﻿/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

using System;
using System.Diagnostics;
using System.IO;
using System.Threading;
using System.Runtime.InteropServices;
using System.Windows.Forms;

namespace MediaLab.TopSurf
{
	public class TOPSURF_DLL
    {
		// prepare external functions
		// Note: we did not compile the C++ dll with 'extern "C"', as that does not allow multiple functions
		//       that have the same name (but differ in parameters), and as a result the function names are
		//       mangled.
		// Note: we support both the 32- and 64-bit dlls, although we cannot find out which one we should
		//       use until at runtime. we can however select between debug and release mode depending on
		//       how we compile the application.
#if DEBUG
		// 32-bit
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_Initialize@@YA_NHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool InitializeDLL32(int imagedim, int top);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_Terminate@@YAXXZ", CallingConvention = CallingConvention.Cdecl)]
		private static extern void TerminateDLL32();
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_LoadDictionary@@YA_NPBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDictionaryDLL32([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_SaveDictionary@@YA_NPBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDictionaryDLL32([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_CreateDictionary@@YA_NPBDHHHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool CreateDictionaryDLL32([MarshalAs(UnmanagedType.LPStr)]string imagedir, int clusters, int knn, int iterations, int points);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPBDAAPAEAAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL32([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPBEHHAAPAEAAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL32(byte[] pixels, int dimx, int dimy, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_CompareDescriptors@@YAMPBE0W4TOPSURF_SIMILARITY@@@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern float CompareDescriptorsDLL32(IntPtr descriptor1, IntPtr descriptor2, TOPSURF_SIMILARITY similarity);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_LoadDescriptor@@YA_NPBDAAPAEAAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDescriptorDLL32([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_SaveDescriptor@@YA_NPBDPBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDescriptorDLL32([MarshalAs(UnmanagedType.LPStr)]string fname, IntPtr descriptor);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_VisualizeDescriptor@@YA_NPAEHHPBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool VisualizeDescriptorDLL32(byte[] pixels, int dimx, int dimy, IntPtr descriptor);
		[DllImport("topsurf_x32d.dll", EntryPoint = "?TopSurf_ReleaseDescriptor@@YAXPAE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern void ReleaseDescriptorDLL32(IntPtr descriptor);
		// 64-bit
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_Initialize@@YA_NHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool InitializeDLL64(int imagedim, int top);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_Terminate@@YAXXZ", CallingConvention = CallingConvention.Cdecl)]
		private static extern void TerminateDLL64();
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_LoadDictionary@@YA_NPEBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDictionaryDLL64([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_SaveDictionary@@YA_NPEBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDictionaryDLL64([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_CreateDictionary@@YA_NPEBDHHHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool CreateDictionaryDLL64([MarshalAs(UnmanagedType.LPStr)]string imagedir, int clusters, int knn, int iterations, int points);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPEBDAEAPEAEAEAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL64([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPEBEHHAEAPEAEAEAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL64(byte[] pixels, int dimx, int dimy, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_CompareDescriptors@@YAMPEBE0W4TOPSURF_SIMILARITY@@@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern float CompareDescriptorsDLL64(IntPtr descriptor1, IntPtr descriptor2, TOPSURF_SIMILARITY similarity);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_LoadDescriptor@@YA_NPEBDAEAPEAEAEAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDescriptorDLL64([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_SaveDescriptor@@YA_NPEBDPEBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDescriptorDLL64([MarshalAs(UnmanagedType.LPStr)]string fname, IntPtr descriptor);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_VisualizeDescriptor@@YA_NPEAEHHPEBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool VisualizeDescriptorDLL64(byte[] pixels, int dimx, int dimy, IntPtr descriptor);
		[DllImport("topsurf_x64d.dll", EntryPoint = "?TopSurf_ReleaseDescriptor@@YAXPEAE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern void ReleaseDescriptorDLL64(IntPtr descriptor);
#else
		// 32-bit
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_Initialize@@YA_NHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool InitializeDLL32(int imagedim, int top);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_Terminate@@YAXXZ", CallingConvention = CallingConvention.Cdecl)]
		private static extern void TerminateDLL32();
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_LoadDictionary@@YA_NPBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDictionaryDLL32([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_SaveDictionary@@YA_NPBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDictionaryDLL32([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_CreateDictionary@@YA_NPBDHHHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool CreateDictionaryDLL32([MarshalAs(UnmanagedType.LPStr)]string imagedir, int clusters, int knn, int iterations, int points);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPBDAAPAEAAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL32([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPBEHHAAPAEAAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL32(byte[] pixels, int dimx, int dimy, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_CompareDescriptors@@YAMPBE0W4TOPSURF_SIMILARITY@@@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern float CompareDescriptorsDLL32(IntPtr descriptor1, IntPtr descriptor2, TOPSURF_SIMILARITY similarity);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_LoadDescriptor@@YA_NPBDAAPAEAAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDescriptorDLL32([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_SaveDescriptor@@YA_NPBDPBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDescriptorDLL32([MarshalAs(UnmanagedType.LPStr)]string fname, IntPtr descriptor);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_VisualizeDescriptor@@YA_NPAEHHPBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool VisualizeDescriptorDLL32(byte[] pixels, int dimx, int dimy, IntPtr descriptor);
		[DllImport("topsurf_x32.dll", EntryPoint = "?TopSurf_ReleaseDescriptor@@YAXPAE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern void ReleaseDescriptorDLL32(IntPtr descriptor);
		// 64-bit
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_Initialize@@YA_NHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool InitializeDLL64(int imagedim, int top);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_Terminate@@YAXXZ", CallingConvention = CallingConvention.Cdecl)]
		private static extern void TerminateDLL64();
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_LoadDictionary@@YA_NPEBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDictionaryDLL64([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_SaveDictionary@@YA_NPEBD@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDictionaryDLL64([MarshalAs(UnmanagedType.LPStr)]string dictionarydir);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_CreateDictionary@@YA_NPEBDHHHH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool CreateDictionaryDLL64([MarshalAs(UnmanagedType.LPStr)]string imagedir, int clusters, int knn, int iterations, int points);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPEBDAEAPEAEAEAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL64([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_ExtractDescriptor@@YA_NPEBEHHAEAPEAEAEAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool ExtractDescriptorDLL64(byte[] pixels, int dimx, int dimy, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_CompareDescriptors@@YAMPEBE0W4TOPSURF_SIMILARITY@@@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern float CompareDescriptorsDLL64(IntPtr descriptor1, IntPtr descriptor2, TOPSURF_SIMILARITY similarity);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_LoadDescriptor@@YA_NPEBDAEAPEAEAEAH@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool LoadDescriptorDLL64([MarshalAs(UnmanagedType.LPStr)]string fname, out IntPtr descriptor, out int length);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_SaveDescriptor@@YA_NPEBDPEBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool SaveDescriptorDLL64([MarshalAs(UnmanagedType.LPStr)]string fname, IntPtr descriptor);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_VisualizeDescriptor@@YA_NPEAEHHPEBE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern bool VisualizeDescriptorDLL64(byte[] pixels, int dimx, int dimy, IntPtr descriptor);
		[DllImport("topsurf_x64.dll", EntryPoint = "?TopSurf_ReleaseDescriptor@@YAXPEAE@Z", CallingConvention = CallingConvention.Cdecl)]
		private static extern void ReleaseDescriptorDLL64(IntPtr descriptor);
#endif

		// the library prints information to stdout and errors to stderr
		// that we would like to capture so we can display them to the user
		TOPSURF_STD m_stdcapture;

		public TOPSURF_DLL(Logger logger)
		{
			// prepare the capture object
			m_stdcapture = new TOPSURF_STD(logger);
		}

		~TOPSURF_DLL()
		{
			// dispose of the capture object
			m_stdcapture.Dispose();
		}

		public static bool Exists()
		{
			// check if the DLL exists
			bool success = false;
			switch (IntPtr.Size)
			{
				case 4:
#if DEBUG
					success = File.Exists("topsurf_x32d.dll");
#else
					success = File.Exists("topsurf_x32.dll");
#endif
					break;
				case 8:
#if DEBUG
					success = File.Exists("topsurf_x64d.dll");
#else
					success = File.Exists("topsurf_x64.dll");
#endif
					break;
			}
			return success;
		}

		public bool Initialize(int imagedim, int top, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// we use a thread to keep the GUI responsive during the long process
			// of creating a dictionary
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = InitializeDLL32(imagedim, top); break;
						case 8: success = InitializeDLL64(imagedim, top); break;
					}
				}
				catch (Exception ex) 
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool Terminate(out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: TerminateDLL32(); success = true; break;
						case 8: TerminateDLL64(); success = true; break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool LoadDictionary(string dictionarydir, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = LoadDictionaryDLL32(dictionarydir); break;
						case 8: success = LoadDictionaryDLL64(dictionarydir); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool SaveDictionary(string dictionarydir, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = SaveDictionaryDLL32(dictionarydir); break;
						case 8: success = SaveDictionaryDLL64(dictionarydir); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool CreateDictionary(string imagedir, int clusters, int knn, int iterations, int points, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = CreateDictionaryDLL32(imagedir, clusters, knn, iterations, points); break;
						case 8: success = CreateDictionaryDLL64(imagedir, clusters, knn, iterations, points); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool ExtractDescriptor(string fname, out TOPSURF_CXX descriptor, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			// Note: we have to use a local variable to pass to the DLL, because
			//       C# is not able to pass the function variable directly
			TOPSURF_CXX d = new TOPSURF_CXX();
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = ExtractDescriptorDLL32(fname, out d.data, out d.length); break;
						case 8: success = ExtractDescriptorDLL64(fname, out d.data, out d.length); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			descriptor = d;
			error = e;
			return success;
		}

		public bool ExtractDescriptor(byte[] pixels, int dimx, int dimy, out TOPSURF_CXX descriptor, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			TOPSURF_CXX d = new TOPSURF_CXX();
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = ExtractDescriptorDLL32(pixels, dimx, dimy, out d.data, out d.length); break;
						case 8: success = ExtractDescriptorDLL64(pixels, dimx, dimy, out d.data, out d.length); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			descriptor = d;
			error = e;
			return success;
		}

		public bool CompareDescriptors(TOPSURF_CXX descriptor1, TOPSURF_CXX descriptor2, TOPSURF_SIMILARITY similarity, out float distance, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			float d = float.MaxValue;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: d = CompareDescriptorsDLL32(descriptor1.data, descriptor2.data, similarity); success = true; break;
						case 8: d = CompareDescriptorsDLL64(descriptor1.data, descriptor2.data, similarity); success = true; break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			distance = d;
			error = e;
			return success;
		}

		public bool LoadDescriptor(string fname, out TOPSURF_CXX descriptor, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			TOPSURF_CXX d = new TOPSURF_CXX();
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = LoadDescriptorDLL32(fname, out d.data, out d.length); break;
						case 8: success = LoadDescriptorDLL64(fname, out d.data, out d.length); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			descriptor = d;
			error = e;
			return success;
		}

		public bool SaveDescriptor(string fname, TOPSURF_CXX descriptor, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = SaveDescriptorDLL32(fname, descriptor.data); break;
						case 8: success = SaveDescriptorDLL64(fname, descriptor.data); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool VisualizeDescriptor(byte[] pixels, int dimx, int dimy, TOPSURF_CXX descriptor, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: success = VisualizeDescriptorDLL32(pixels, dimx, dimy, descriptor.data); break;
						case 8: success = VisualizeDescriptorDLL64(pixels, dimx, dimy, descriptor.data); break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}

		public bool ReleaseDescriptor(TOPSURF_CXX descriptor, out string error)
		{
			// reset the capture object
			m_stdcapture.Reset();
			// call the DLL function from within the thread
			bool success = false;
			string e = null;
			ThreadStart starter = delegate
			{
				try
				{
					switch (IntPtr.Size)
					{
						case 4: ReleaseDescriptorDLL32(descriptor.data); success = true; break;
						case 8: ReleaseDescriptorDLL64(descriptor.data); success = true; break;
					}
				}
				catch (Exception ex)
				{
					success = false;
					e = ex.Message;
				}
			};
			Thread thread = new Thread(starter);
			thread.Start();
			while (thread.IsAlive)
			{
				Application.DoEvents();
				Thread.Sleep(100);
			}
			thread.Join();
			error = e;
			return success;
		}
	}
}