﻿using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;

namespace MediaLab.TopSurf
{
	class TOPSURF_STD : IDisposable
	{
		// the library prints information to stdout and errors to stderr
		// that we would like to capture so we can display them to the user
		[DllImport("Kernel32.dll", SetLastError = true)]
		public static extern int SetStdHandle(int device, IntPtr handle);

		Logger m_logger;
		string m_stdoutName = null;
		string m_stderrName = null;
		TextWriter m_stdoutOriginal = null;
		TextWriter m_stderrOriginal = null;
		FileStream m_stdoutStream = null;
		FileStream m_stderrStream = null;
		StreamWriter m_stdoutWriter = null;
		StreamWriter m_stderrWriter = null;
		int m_stdoutLines;
		int m_stderrLines;
		FileSystemWatcher m_fileWatcher = null;
		bool m_redirect;

		public TOPSURF_STD(Logger logger)
		{
			// set logger
			m_logger = logger;
			// get unique identifier
			// Note: just in case multiple GUIs are run simultaneously
			int uniqueid = Process.GetCurrentProcess().Id;
			// set stdout
			try
			{
				m_stdoutName = Path.Combine(Path.GetTempPath(), "stdout" + uniqueid + ".txt");
				m_stdoutStream = new FileStream(m_stdoutName, FileMode.Create, FileAccess.Write, FileShare.Read);
				m_stdoutWriter = new StreamWriter(m_stdoutStream);
				m_stdoutWriter.AutoFlush = true;
				m_stdoutOriginal = Console.Out;
				Console.SetOut(m_stdoutWriter);
				IntPtr stdoutHandle = m_stdoutStream.SafeFileHandle.DangerousGetHandle();
				SetStdHandle(-11, stdoutHandle);
				m_stdoutLines = 0;
				// set stderr
				m_stderrName = Path.Combine(Path.GetTempPath(), "stderr" + uniqueid + ".txt");
				m_stderrStream = new FileStream(m_stderrName, FileMode.Create, FileAccess.Write, FileShare.Read);
				m_stderrWriter = new StreamWriter(m_stderrStream);
				m_stderrWriter.AutoFlush = true;
				m_stderrOriginal = Console.Error;
				Console.SetError(m_stderrWriter);
				IntPtr stderrHandle = m_stderrStream.SafeFileHandle.DangerousGetHandle();
				SetStdHandle(-12, stderrHandle);
				m_stderrLines = 0;
				// initialize the file system watcher that we use to keep track of changes
				// to the stdout and stderr files
				m_fileWatcher = new FileSystemWatcher(Path.GetTempPath(), "*.txt");
				m_fileWatcher.Changed += new FileSystemEventHandler(fileWatcher_Changed);
				m_fileWatcher.EnableRaisingEvents = true;
				// indicate the redirection was successful
				m_redirect = true;
			}
			catch (Exception ex)
			{
				Console.Error.WriteLine(ex.Message);
				m_stdoutStream = null;
				m_stdoutWriter = null;
				m_stderrStream = null;
				m_stderrWriter = null;
				m_fileWatcher = null;
				m_redirect = false;
			}
		}

		~TOPSURF_STD()
		{
			// finalizer calls Dispose(false)
			Dispose(false);
		}

		public void Dispose()
		{
			// Dispose() calls Dispose(true)
			Dispose(true);
			GC.SuppressFinalize(this);
		}
		
		protected virtual void Dispose(bool disposing)
		{
			// release unmanaged data
			if (disposing && m_redirect)
			{
				m_redirect = false;
				m_fileWatcher.EnableRaisingEvents = false;
				m_fileWatcher.Dispose();
				m_fileWatcher = null;
				// Note: only the streamwriter needs to be closed, as
				//       it automatically closes the filestream too
				try { Console.SetOut(m_stdoutOriginal); }
				catch { }
				finally { m_stdoutOriginal = null; }
				m_stdoutStream.Close();
				m_stdoutStream = null;
				m_stdoutWriter = null;
				try { File.Delete(m_stdoutName); }
				catch { }
				finally { m_stdoutName = null; }
				try { Console.SetOut(m_stderrOriginal); }
				catch { }
				finally { m_stderrOriginal = null; }
				m_stderrStream.Close();
				m_stderrStream = null;
				m_stderrWriter = null;
				try { File.Delete(m_stderrName); }
				catch { }
				finally { m_stderrName = null; }
			}
		}

		void fileWatcher_Changed(object sender, FileSystemEventArgs e)
		{
			if (e.FullPath.Equals(m_stdoutName))
			{
				try
				{
					// open the file for concurrent read access
					FileStream stream = new FileStream(e.FullPath, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
					TextReader reader = new StreamReader(stream);
					// read in all lines up to where we have already read
					for (int i = 0; i < m_stdoutLines; i++)
					{
						if (reader.ReadLine() == null)
							break;
					}
					// read in any new lines
					while (true)
					{
						string s = reader.ReadLine();
						if (s == null)
							break;
						m_logger.WriteNormal(s);
						m_stdoutLines++;
					}
					reader.Close();
				}
				catch (Exception ex)
				{
					Debug.WriteLine(ex.Message);
					return;
				}
			}
			else if (e.FullPath.Equals(m_stderrName))
			{
				try
				{
					// open the file for concurrent read access
					FileStream stream = new FileStream(e.FullPath, FileMode.Open, FileAccess.Read, FileShare.ReadWrite);
					TextReader reader = new StreamReader(stream);
					// read in all lines up to where we have already read
					for (int i = 0; i < m_stderrLines; i++)
					{
						if (reader.ReadLine() == null)
							break;
					}
					// read in any new lines
					while (true)
					{
						string s = reader.ReadLine();
						if (s == null)
							break;
						m_logger.WriteError(s);
						m_stderrLines++;
					}
					reader.Close();
				}
				catch (Exception ex)
				{
					Debug.WriteLine(ex.Message);
					return;
				}
			}
		}

		public void Reset()
		{
			if (m_redirect)
			{
				m_fileWatcher.EnableRaisingEvents = false;
				m_stdoutStream.SetLength(0);
				m_stdoutLines = 0;
				m_stderrStream.SetLength(0);
				m_stderrLines = 0;
				m_fileWatcher.EnableRaisingEvents = true;
			}
		}
	}
}
