/*	TOP-SURF: a visual words toolkit
	Copyright (C) 2010 LIACS Media Lab, Leiden University,
	                   Bart Thomee (bthomee@liacs.nl),
					   Erwin M. Bakker (erwin@liacs.nl)	and
					   Michael S. Lew (mlew@liacs.nl).

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.
    See http://www.gnu.org/licenses/gpl.html for the full license.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

	In addition, this work is covered under the Creative Commons
	Attribution license version 3.
    See http://creativecommons.org/licenses/by/3.0/ for the full license.
*/

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Text;
using System.Windows.Forms;

namespace MediaLab.TopSurf
{
    public partial class VisualizeForm : Form
    {
		Bitmap m_bitmap;
		TOPSURF_DESCRIPTOR m_descriptor;

        public VisualizeForm(TOPSURF_GUIDATA data)
        {
            InitializeComponent();
			// set the title
			if (!string.IsNullOrEmpty(data.Name))
				this.Text = data.Name;
            // set the image as the background image and resize the form to fit
			if (data.Visualize == null)
				throw new ArgumentException();
			Rectangle workingArea = Screen.GetWorkingArea(this);
			if (data.Visualize.Width > workingArea.Width || data.Visualize.Height > workingArea.Height)
			{
				if (data.Visualize.Width * workingArea.Height / workingArea.Width < data.Visualize.Height)
					this.ClientSize = new Size(data.Visualize.Width / (data.Visualize.Height / workingArea.Height), workingArea.Height);
				else
					this.ClientSize = new Size(workingArea.Width, data.Visualize.Height / (data.Visualize.Width / workingArea.Width));
			}
			else
				this.ClientSize = new Size(data.Visualize.Width, data.Visualize.Height);
			m_bitmap = new Bitmap(data.Visualize);
			this.BackgroundImage = m_bitmap;
			// convert the descriptor
			m_descriptor = TOPSURF_DESCRIPTOR.ConvertDescriptor(data.Descriptor);
        }

		private void VisualizeForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (this.BackgroundImage != null)
				this.BackgroundImage.Dispose();
		}

		private void VisualizeForm_MouseClick(object sender, MouseEventArgs e)
		{
			if (e.Button == MouseButtons.Right)
			{
				// show context menu
				contextMenuStrip.Show(this, e.Location);
			}
		}

		private void VisualizeForm_MouseMove(object sender, MouseEventArgs e)
		{
			double factorx = (double)this.ClientSize.Width;
			double factory = (double)this.ClientSize.Height;
			double factors = (factorx + factory) / 2;
			// convert the mouse coordinates so their origin is the bottom-left
			// hand corner of the image
			double x = (double)e.X;
			double y = factory - (double)e.Y - 1;
			// determine the closest visual word to the mouse
			double distance = double.MaxValue;
			int identifier = -1;
			foreach (TOPSURF_VISUALWORD vw in m_descriptor.visualword)
			{
				foreach (TOPSURF_LOCATION l in vw.location)
				{
					// start with a rectangle around the origin with size of the scale
					double x1 = -l.scale * factors;
					double y1 = -l.scale * factors;
					double x2 = -l.scale * factors;
					double y2 = l.scale * factors;
					double x3 = l.scale * factors;
					double y3 = l.scale * factors;
					double x4 = l.scale * factors;
					double y4 = -l.scale * factors;
					// rotate the rectangle around the origin using the orientation
					double x11 = x1 * Math.Cos(l.orientation) - y1 * Math.Sin(l.orientation);
					double y11 = x1 * Math.Sin(l.orientation) + y1 * Math.Cos(l.orientation);
					double x22 = x2 * Math.Cos(l.orientation) - y2 * Math.Sin(l.orientation);
					double y22 = x2 * Math.Sin(l.orientation) + y2 * Math.Cos(l.orientation);
					double x33 = x3 * Math.Cos(l.orientation) - y3 * Math.Sin(l.orientation);
					double y33 = x3 * Math.Sin(l.orientation) + y3 * Math.Cos(l.orientation);
					double x44 = x4 * Math.Cos(l.orientation) - y4 * Math.Sin(l.orientation);
					double y44 = x4 * Math.Sin(l.orientation) + y4 * Math.Cos(l.orientation);
					// move the point to the location it was detected
					double x111 = x11 + l.x * factorx;
					double y111 = y11 + l.y * factory;
					double x222 = x22 + l.x * factorx;
					double y222 = y22 + l.y * factory;
					double x333 = x33 + l.x * factorx;
					double y333 = y33 + l.y * factory;
					double x444 = x44 + l.x * factorx;
					double y444 = y44 + l.y * factory;
					// check if the mouse is inside the rectangle
					// Note: assuming the edges are drawn in a clockwise direction,
					//       it comes down to checking if the point lies to the right
					//       or each edge
					double ex = x222 - x111;
					double ey = y222 - y111;
					double fx = x444 - x111;
					double fy = y444 - y111;
					if ((x - x111) * ex + (y - y111) * ey < 0.0 ||
						(x - x222) * ex + (y - y222) * ey > 0.0 ||
						(x - x111) * fx + (y - y111) * fy < 0.0 ||
						(x - x444) * fx + (y - y444) * fy > 0.0)
						continue;
					// determine the distance between the mouse and each of the points
					double d1 = Math.Sqrt((x - x111) * (x - x111) + (y - y111) * (y - y111));
					double d2 = Math.Sqrt((x - x222) * (x - x222) + (y - y222) * (y - y222));
					double d3 = Math.Sqrt((x - x333) * (x - x333) + (y - y333) * (y - y333));
					double d4 = Math.Sqrt((x - x444) * (x - x444) + (y - y444) * (y - y444));
					double d = Math.Min(d1, d2);
					d = Math.Min(d, d3);
					d = Math.Min(d, d4);
					if (d < distance)
					{
						distance = d;
						identifier = vw.identifier;
					}
				}
			}
			// show of hide tooltip
			if (identifier != -1)
			{
				// place the tooltip either above or below the cursor, depending on
				// the location of the cursor
				int offsetY;
				if (e.Y < this.ClientSize.Height / 2)
					offsetY = SystemInformation.CaptionHeight * 2;
				else
					offsetY = 0;
				toolTip.Show("identifier=" + identifier.ToString(), this, e.X, e.Y + offsetY);
			}
			else
				toolTip.Hide(this);
		}

		private void VisualizeForm_KeyUp(object sender, KeyEventArgs e)
		{
			if (e.KeyCode == Keys.S)
				SaveImage();
		}

		private void saveImageMenuItem_Click(object sender, EventArgs e)
		{
			SaveImage();
		}

		private void SaveImage()
		{
			// save the image to disk
			using (SaveFileDialog sbd = new SaveFileDialog())
			{
				sbd.InitialDirectory = Directory.GetCurrentDirectory();
				sbd.RestoreDirectory = true;
				sbd.Title = "Save image";
				sbd.Filter = "JPG File|*.jpg|PNG File|*.png";
				if (sbd.ShowDialog() != DialogResult.OK)
					return;
				try
				{
					// determine image format
					ImageFormat imageFormat;
					if (sbd.FilterIndex == 1)
						imageFormat = ImageFormat.Jpeg;
					else
						imageFormat = ImageFormat.Png;
					// save the image
					Stream stream = sbd.OpenFile();
					if (stream != null)
					{
						this.BackgroundImage.Save(stream, imageFormat);
						stream.Close();
					}
				}
				catch (Exception ex)
				{
					MessageBox.Show("Could not save image.\n" + ex.Message, "An error occurred", MessageBoxButtons.OK, MessageBoxIcon.Error);
				}
			}
		}
    }
}